// ---------------------------------------------
// Simple Synchronized FIFO
// Author : Zhu, Shenli
// Version 1 
//----------------------------------------------

module sync_fifo_rtl (
    CLK,
    RESET,
    
    // control signal
    RD,
    WR,
    // data path
    DIN,
    DOUT,
    // status signal
    FULL,
    EMPTY,
    ALMOST_FULL, // one clock ahead of FULL
    ALMOST_EMPTY // one clock ahead of EMPTY
    );
parameter FIFO_WIDTH = 32;
parameter FIFO_DEPTH = 8;
parameter PTR_WIDTH = 3; //i.e. log2(FIFO_DEPTH) 

input CLK;
input RESET;
input RD;
input WR;
input [FIFO_WIDTH-1:0] DIN;
output [FIFO_WIDTH-1:0] DOUT;
reg [FIFO_WIDTH-1:0] DOUT;
output  FULL;
output  EMPTY;
output ALMOST_FULL;
output ALMOST_EMPTY;

reg [FIFO_WIDTH-1:0] mem [FIFO_DEPTH-1:0];
reg [PTR_WIDTH-1:0] wr_minus_rd; // wr_minus_rd is the only pointer in Synchronized FIFO
reg [PTR_WIDTH-1:0] wr_ptr;
reg [PTR_WIDTH-1:0] rd_ptr;

assign FULL = ( wr_minus_rd == (FIFO_DEPTH - 1) );
assign EMPTY = ( wr_minus_rd == 0 );
assign ALMOST_FULL = ( wr_minus_rd == (FIFO_DEPTH - 2) ) |  FULL;
assign ALMOST_EMPTY = ( wr_minus_rd == 1 );

// updata the wr_minus_rd
// ( 0 <= wr_minus_rd < FIFO_DEPTH )
always @ (negedge RESET or posedge CLK) begin
    if ( ~RESET ) begin
        wr_minus_rd <= #1 0; 
    end
    else if ( (WR == 1) && (RD == 0) ) begin
        wr_minus_rd <= #1 wr_minus_rd + 1;
    end
    else if ( (WR == 0) && (RD == 1) ) begin
        wr_minus_rd <= #1 wr_minus_rd - 1;
    end
    else begin
        wr_minus_rd <= #1 wr_minus_rd;
    end
end

// update wr_ptr
always @ (negedge RESET or posedge CLK) begin
    if ( ~RESET ) begin
        wr_ptr <= #1 0; 
    end
    else if (WR == 1)begin
        wr_ptr <= #1 wr_ptr + 1;
    end
end

// update rd_ptr
always @ (negedge RESET or posedge CLK) begin
    if ( ~RESET ) begin
        rd_ptr <= #1 0; 
    end
    else if (RD == 1)begin
        rd_ptr <= #1 rd_ptr + 1;
    end
end


// read DIN
integer mem_cnt; 
always @ (negedge RESET or posedge CLK) begin
    if ( ~RESET ) begin
        // initialize mem to all zero
        for (mem_cnt = 0; mem_cnt < FIFO_DEPTH; mem_cnt = mem_cnt + 1) begin
            mem[mem_cnt] <= #1 0;
        end
    end
    else if (WR == 1) begin
        mem[wr_ptr] <= #1 DIN;
    end
end

// write DOUT
always @ (negedge RESET or posedge CLK) begin
    if ( ~RESET ) begin
        DOUT <= #1 0;
    end
    else if (RD == 1) begin
        DOUT <= #1 mem[rd_ptr];
    end
end

endmodule

